#include "TCP.h"
#include "utility/w5500.h"
#include "utility/socket.h"

std::vector<EthernetMaketClient*> EthernetMaketClient::instances;


EthernetMaketClient::EthernetMaketClient() {
    _sock = MAX_SOCK_NUM;
    instances.push_back(this);
}

EthernetMaketClient::EthernetMaketClient(uint8_t sock) {
    _sock = sock;
    instances.push_back(this);
}

EthernetMaketClient::~EthernetMaketClient() {
    instances.erase(std::remove(instances.begin(), instances.end(), this), instances.end());
}

void EthernetMaketClient::tick() {
    for (EthernetMaketClient* client : instances) {
        client->tick_();
    }
}

ConnectionStatusSimple EthernetMaketClient::connectNonBlock(IPAddress ip, uint16_t port) {
    dstIP = ip;
    dstPort = port;
  if (_sock != MAX_SOCK_NUM) {
    // Serial.println(F("connectNonBlock"));
    uint8_t status_ = status();
    switch (status_) {
      case SnSR::CLOSED:
        // Serial.println(F("CLOSED"));
        _sock = MAX_SOCK_NUM;
        return CONNECT_FAIL;
        break;
      case SnSR::ESTABLISHED:
        // Serial.println(F("ESTABLISHED"));
        return CONNECT_SUCCESS;
        break;
      default:
        // Serial.print(F("status "));
        // Serial.println(status_,HEX);
        return CONNECT_CONNECTING;
        break;
    }
  }
  for (int i = 0; i < MAX_SOCK_NUM; i++) {
    uint8_t s = w5500.readSnSR(i);
    if (s == SnSR::CLOSED || s == SnSR::FIN_WAIT || s == SnSR::CLOSE_WAIT) {
      _sock = i;
      break;
    }
  }

  if (_sock == MAX_SOCK_NUM)
    return CONNECT_FAIL;

  _srcport++;
  if (_srcport == 0) _srcport = 1024;
  socket(_sock, SnMR::TCP, _srcport, 0);

  if (!::connect(_sock, rawIPAddress(ip), port)) {
    _sock = MAX_SOCK_NUM;
    return CONNECT_FAIL;
  }
  
  return CONNECT_CONNECTING;
}

bool EthernetMaketClient::isConnected() const {
    return connectStatus == CONNECT_CONNECTED;
}

void EthernetMaketClient::disconnect(){
  if (_sock == MAX_SOCK_NUM)
    return;
    ::disconnect(_sock);
}

void EthernetMaketClient::close(){
  if (_sock == MAX_SOCK_NUM)
    return;
    ::close(_sock);
    _sock = MAX_SOCK_NUM;
}

void EthernetMaketClient::tick_(){
  // Подключаемся к серверу
    switch (connectStatus)
    {
    case CONNECT_START:
        startConnection = millis();
        connectStatus = CONNECT_CONNECTING;
        close();
        Serial.println("\n\nConnecting...");
        break;
    case CONNECT_CONNECTING:
        if (millis() - startConnection >= connectionTimeout)
        {
            connectStatus = CONNECT_FAIL;
            break;
        }
        break;
    case CONNECT_FAIL:
        Serial.print("Failed to connect to server in ");
        Serial.println(isNonBlocking ? "non-blocking mode" : "blocking mode");
        connectStatus = CONNECT_STOP_START;
        break;
    case CONNECT_SUCCESS:
        Serial.print("Connected to server in ");
        Serial.println(isNonBlocking ? "non-blocking mode" : "blocking mode");
        lastActivityTime = millis();
        connectStatus = CONNECT_CONNECTED;

            lastActivityTime = millis(); // Обновляем время последней активности

        break;
    case CONNECT_CONNECTED:

        //! Передача пакетов
        dataWrite();
        //! Передача завершена

        if (!connected())
        {
            Serial.println("Client disconnected");
            connectStatus = CONNECT_STOP_START;
        }

        // Закрытие соединения по таймеру неактивности
        if (connected() && (millis() - lastActivityTime > timeout))
        {
            Serial.println("Connection closed due to inactivity");
            connectStatus = CONNECT_STOP_START;
        }
        break;

    case CONNECT_STOP_START:
        dstIP = IPAddress(0,0,0,0);
        dstPort = 0;
        disconnect();
        stopStartTime = millis();
        connectStatus = CONNECT_STOP_WAITING;
    break;

    case CONNECT_STOP_WAITING:
        if (millis() - stopStartTime >= stopTimeout) {
            close(); // Окончательное закрытие сокета после таймаута
            connectStatus = (dataSize > 0) ? CONNECT_START : CONNECT_IDLE;
        } else {
            // Проверяем состояние сокета для завершения разрыва
            uint8_t status_ = status();
            if (status_ == SnSR::CLOSED || status_ == SnSR::CLOSE_WAIT) {
                close(); // Закрываем при достижении состояния
                connectStatus = (dataSize > 0) ? CONNECT_START : CONNECT_IDLE;
            }
        }
        break;

    case CONNECT_IDLE:
        if(dataSize){
            connectStatus = CONNECT_START;
        }
        break;

    default:
        break;
    }
//********************************************************
    switch (connectStatus)
    {
    case CONNECT_START:
    case CONNECT_CONNECTING:
        if(dstIP == IPAddress(0,0,0,0) || dstPort == 0)
            break;

        if (isNonBlocking)
        {
            connectStatus = connectNonBlock(dstIP, dstPort);
        }
        else
        {
            connectStatus = (ConnectionStatusSimple)connect(dstIP, dstPort);
        }
        break;

    default:
        break;
    }
//********************************************************

    //? Проверка наличия доступных данных от сервера
    if (available() > 0)
    {
        byte response[16];
        int len = read(response, sizeof(response));

        if (len == 16)
        {
            // Проверка контрольной суммы
            uint8_t checksum = 0;
            for (int i = 0; i < 15; i++)
            {
                checksum += response[i];
            }

            Serial.print("RX: ");
            for (int i = 0; i < 16; i++)
            {
                Serial.print(response[i], HEX);
                Serial.print(" ");
            }

            if (response[15] == checksum)
            {
                Serial.println("Checksum is OK");
            }
            else
            {
                Serial.println("Checksum is ERROR");
            }
            lastActivityTime = millis(); // Обновляем время последней активности
        }
        else
        {
            Serial.println("Received an incomplete response");
        }
    }
    //? -------------------------------------------------
}

bool EthernetMaketClient::dataWrite(){
    bool ret = false;
    if(dataSize) {
        lastActivityTime = millis();
        if(write((uint8_t *)&dataPtr, dataSize)){
            dataPtr = nullptr; 
            dataSize = 0; // Очищаем размер данных
            ret = true;
        }
        lastActivityTime = millis();
    }
    return ret;
}

bool EthernetMaketClient::dataWrite(uint8_t* data, uint16_t dataSize, bool override) {
    if (dataSize == 0 || override || (dataPtr == nullptr && this->dataSize == 0)) {
        this->dataPtr = data;
        this->dataSize = dataSize;
        return dataWrite();
    } else {
        Serial.println("Data buffer is not empty. Write operation is ignored.");
        return false;
    }
}


bool EthernetMaketClient::send(uint8_t* data, uint16_t dataSize, IPAddress ip, uint16_t port) {
    this->dataPtr = data;
    this->dataSize = dataSize;

    switch (connectStatus) {
        case CONNECT_CONNECTED:
            if(dataWrite()){
                onSendSuccess(); 
            }
            // Если подключено, отправляем данные
            return true;

        case CONNECT_STOP_WAITING:
        case CONNECT_STOP_START:
            return false; // Ожидаем завершения разрыва

        case CONNECT_FAIL:
        case CONNECT_IDLE:
            connectNonBlock(ip, port);
            connectStatus = CONNECT_START; // Начинаем новое подключение
            return false; // Подключение инициировано

        default:
            return false; // Пока ждём подключения
    }
}

void EthernetMaketClient::setOnSendSuccess(const std::function<void()>& callback) {
    onSendSuccess = callback;
}

void EthernetMaketClient::resetOnSendSuccess() {
    onSendSuccess = []() {};
}
